/*

CLMTSMeshFileLoader.cpp

LMTSMeshFileLoader
Written by Jonas Petersen (a.k.a. jox)

Version 1.3 (15 Feburary 2005)

Get the latest version here: http://development.mindfloaters.de/

This is an addon class for the Irrlicht engine by Nikolaus Gebhardt (http://irrlicht.sourceforge.net).
With this release the Irrlicht engine is at version 0.6

This class allows loading meshes with lightmaps (*.lmts + *.tga files) that were created
using Pulsar LMTools by Lord Trancos (http://www.geocities.com/dxlab/index_en.html)

Notes:
- This version does not recognice/support user data in the *.lmts files.
- The mesh must have at least one texture and one lightmap.
- The lightmap TGA's generated by LMTools doesn't work in Irrlicht for some reason (the
  lightmaps look messed up). Opening and resaving them in a graphics app will solve
  the problem (tested only with Photoshop).


License:
--------

It's free. You are encouraged to give me credit if you use it in your software.


Usage:
------

I assume you know how to generate meshes with lightmaps using LMTools.

Put the 2 files CLMTSMeshFileLoader.h and CLMTSMeshFileLoader.cpp in your app path.

Make sure the textures and lightmap TGA's can be found by the app.

Here's some example code:

//Include the header file:

	#include "CLMTSMeshFileLoader.h"

//Preparation things you might have in your app already:

	scene::ISceneManager* smgr = device->getSceneManager();
	video::IVideoDriver* driver = device->getVideoDriver();
	io::IFileSystem* fs = device->getFileSystem();

//I recommend setting one of these flags for nicer results:
  
	driver->setTextureCreationFlag(video::ETCF_ALWAYS_32_BIT,true); 
	// or
	driver->setTextureCreationFlag(video::ETCF_OPTIMIZED_FOR_QUALITY, true);

//Register the mesh loader (the important part!):

	scene::CLMTSMeshFileLoader* lmts_loader = new scene::CLMTSMeshFileLoader(fs, driver);
	smgr->addExternalMeshLoader(lmts_loader);

// Optionally set path where the textures can be found (path MUST end with a backslash):

	lmts_loader->setTexturePath("some_dir\\");
	// or
	lmts_loader->setTexturePath("C:\\path\\to\\textures\\");

//Load a mesh:

	scene::IAnimatedMesh* mesh = smgr->getMesh("some_dir\\some_mesh.lmts");

//Add the mesh to the scene:

	scene::IAnimatedMeshSceneNode* node = smgr->addAnimatedMeshSceneNode( mesh );

That's it! (hopefully)

  
Version History:
----------------

Version 1.3 - 15 February 2005
- Fixed bug that prevented loading more than one different lmts files.
- Removed unnecessary "#include <os.h>".
- Added "std::" in front of "cerr". This was necessary for Visual Studio .NET,
  I hope it's not disturbing other compilers.
- Added warning message when a texture can not be loaded.
- Changed the documentation a bit (minor).

Version 1.2
- To avoid confusion I skipped version 1.2 because the website was offering
version 1.2 even though it was only version 1.1. Sorry about that.

Version 1.1 - 29 July 2004
- Added setTexturePath() function
- Minor improvements

Version 1.0 - 29 July 2004
- Initial release


*/
//////////////////////////////////////////////////////////////////////

#include <string.h>
#include <SMeshBufferLightMap.h>
#include <SAnimatedMesh.h>
#include <SMeshBuffer.h>
#include <irrString.h>
#include <stdio.h>
#include "CLMTSMeshFileLoader.h"

namespace irr
{
namespace scene
{

CLMTSMeshFileLoader::CLMTSMeshFileLoader(io::IFileSystem* fs, video::IVideoDriver* driver)
: Textures(0), Subsets(0), Triangles(0), NumTextures(0), NumLightMaps(0),
  TextureIDs(0), Mesh(0), Driver(driver), FileSystem(fs)
{
	if (Driver)
		Driver->grab();

	if (FileSystem)
		FileSystem->grab();
}

CLMTSMeshFileLoader::~CLMTSMeshFileLoader() {
	if (Textures)
		delete [] Textures;

	if (Subsets)
		delete [] Subsets;

	if (Triangles)
		delete [] Triangles;

	if (TextureIDs)
		delete [] TextureIDs;

	if (Mesh)
		Mesh->drop();

	if (Driver)
		Driver->drop();

	if (FileSystem)
		FileSystem->drop();

}

bool CLMTSMeshFileLoader::isALoadableFileExtension(const c8* filename) {
	return strstr(filename, ".lmts") != 0;
}


IAnimatedMesh* CLMTSMeshFileLoader::createMesh(irr::io::IReadFile* file) {

	u32 i;
	u32 id;

	file->seek(0);

	// HEADER

	file->read(&Header, sizeof(SLMTSHeader));
	if (Header.MagicID != 0x53544D4C) { // "LMTS"
		std::cerr << "LMTS LOADER ERROR: wrong header magic id!\n";
		return 0;
	}

	// TEXTURES

	file->read(&id, sizeof(u32));
	if (id != 0x54584554) { // "TEXT"
		std::cerr << "LMTS LOADER ERROR: wrong texture magic id!\n";
		return 0;
	}

	Textures = new SLMTSTextureInfoEntry[Header.TextureCount];
	TextureIDs = new u16[Header.TextureCount];

	NumLightMaps = NumTextures = 0;

	for (i=0; i<Header.TextureCount; i++) {
		file->read(&Textures[i], sizeof(SLMTSTextureInfoEntry));
		if (Textures[i].Flags & 1) {
			TextureIDs[i] = NumLightMaps;
			NumLightMaps++;
		} else {
			TextureIDs[i] = NumTextures;
			NumTextures++;
		}

	}

	// SUBSETS

	file->read(&id, sizeof(u32));
	if (id != 0x53425553) { // "SUBS"
		std::cerr << "LMTS LOADER ERROR: wrong subset magic id!\n";
		return 0;
	}

	Subsets = new SLMTSSubsetInfoEntry[Header.SubsetCount];

	for (i=0; i<Header.SubsetCount; i++) {
		file->read(&Subsets[i], sizeof(SLMTSSubsetInfoEntry));
	}

	// TRIANGLES

	file->read(&id, sizeof(u32));
	if (id != 0x53495254) { // "TRIS"
		std::cerr << "LMTS LOADER ERROR: wrong triangle magic id!\n";
		return 0;
	}

	Triangles = new SLMTSTriangleDataEntry[(Header.TriangleCount*3)];

	for (i=0; i<(Header.TriangleCount*3); i++) {
		file->read(&Triangles[i], sizeof(SLMTSTriangleDataEntry));
	}

    /////////////////////////////////////////////////////////////////

	constructMesh();

	loadTextures();

	SAnimatedMesh* am = new SAnimatedMesh();
	am->Type = EAMT_UNKNOWN; // well, unknown to irrlicht

	am->addMesh(Mesh);
	am->recalculateBoundingBox();
	Mesh->drop();
	Mesh = 0;
    return am;
	
}

void CLMTSMeshFileLoader::constructMesh()
{
	s32 i;

	if (Mesh)
		Mesh->drop();

	Mesh = new SMesh();

	for (i=0; i<(NumTextures * NumLightMaps); ++i)
	{
		scene::SMeshBufferLightMap* buffer = new scene::SMeshBufferLightMap();

		buffer->Material.MaterialType = video::EMT_LIGHTMAP; // EMT_LIGHTMAP_M2/EMT_LIGHTMAP_M4 also possible
		buffer->Material.Wireframe = false;
		buffer->Material.Lighting = false;
		buffer->Material.BilinearFilter = true;

		Mesh->addMeshBuffer(buffer);

		buffer->drop();
	}


	for (i=0; i<Header.SubsetCount; i++) {
		u16 textureID = TextureIDs[Subsets[i].TextID1];
		u16 lightmapID = TextureIDs[Subsets[i].TextID2];
		u32 offs = Subsets[i].Offset * 3;

		s32 meshBufferIndex = lightmapID * NumTextures + textureID;
		SMeshBufferLightMap* meshBuffer = ((SMeshBufferLightMap*)Mesh->getMeshBuffer(meshBufferIndex));

		for (u32 sc=0; sc<Subsets[i].Count; sc++) {

			s32 idx = meshBuffer->getVertexCount();
			
			for (s32 vu=0; vu<3; ++vu)
			{
				video::S3DVertex2TCoords currentVertex;
				SLMTSTriangleDataEntry *v = &Triangles[offs+(3*sc)+vu];

				currentVertex.Color.set(255,255,255,255);

				currentVertex.Pos.X = v->X;
				currentVertex.Pos.Y = v->Y;
				currentVertex.Pos.Z =  v->Z;
				currentVertex.TCoords.X = v->U1;
				currentVertex.TCoords.Y = v->V1;
				currentVertex.TCoords2.X = v->U2;
				currentVertex.TCoords2.Y = v->V2;

				meshBuffer->Vertices.push_back(currentVertex);
			}

			meshBuffer->Indices.push_back(idx);
			meshBuffer->Indices.push_back(idx+1);
			meshBuffer->Indices.push_back(idx+2);
		}
	}

	for (u32 j=0; j<Mesh->MeshBuffers.size(); ++j)
		((SMeshBufferLightMap*)Mesh->MeshBuffers[j])->recalculateBoundingBox();

	Mesh->recalculateBoundingBox();
}

void CLMTSMeshFileLoader::setTexturePath(const c8* path) {
	Path = path;
}

void CLMTSMeshFileLoader::loadTextures()
{
	if (!Driver || !FileSystem)
		return;

	core::stringc s;

	// load textures

	core::array<video::ITexture*> tex;
	tex.set_used(NumTextures);

	core::array<video::ITexture*> lig;
	lig.set_used(NumLightMaps);

	s32 t;
	s32 tx_count = 0;
	s32 lm_count = 0;
	for (t=0; t<Header.TextureCount; t++) {
		video::ITexture* tmptex = 0;
		s = Path;
		s.append(Textures[t].Filename);

		if (FileSystem->existFile(s.c_str()))
			tmptex = Driver->getTexture(s.c_str());
		else std::cerr << "LMTS LOADER ERROR: Texture does not exist: " << s.c_str() << "\n";


		if (Textures[t].Flags & 1) {
			lig[lm_count++] = tmptex;
		} else {
			tex[tx_count++] = tmptex;
		}

	}


	// attach textures to materials.

	for (s32 l=0; l<NumLightMaps; ++l)
		for (s32 t=0; t<NumTextures; ++t)
		{
			SMeshBufferLightMap* b = (SMeshBufferLightMap*)Mesh->getMeshBuffer(l*(NumTextures) + t);

			b->Material.Texture2 = lig[l];
			b->Material.Texture1 = tex[t];

			if (!b->Material.Texture2)
				b->Material.MaterialType = video::EMT_SOLID;
		}

	// delete all buffers without geometry in it.

	s32 i = 0;
	while(i < (s32)Mesh->MeshBuffers.size())
	{
		if (Mesh->MeshBuffers[i]->getVertexCount() == 0 ||
			Mesh->MeshBuffers[i]->getIndexCount() == 0 ||
			Mesh->MeshBuffers[i]->getMaterial().Texture1 == 0)
		{
			Mesh->MeshBuffers[i]->drop();
			Mesh->MeshBuffers.erase(i);		
		}
		else
			++i;
	}
}


} // end namespace scene
} // end namespace irr
